#!/usr/bin/env python3
"""Script that integrates everything SCC can do in one executable.

Created so scc-* stuff doesn't polute /usr/bin.
"""

from scc.scripts import main

if __name__ == "__main__":
	main()
